#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <semaphore.h>

#include <string>

// ***********************************************************************
// Moment
// ***********************************************************************

class Moment {
public:
	void now () {
		(void) gettimeofday(&tv,NULL);
	};

	double operator - ( Moment &m ) {
		double result = ((double)  tv.tv_sec - m.tv.tv_sec) * 1e6;
		result += ((double) tv.tv_usec - m.tv.tv_usec);
		return result/1000.0;
	};
private:
	struct timeval tv;
};

// ***********************************************************************
// Thread
// ***********************************************************************

typedef void *(*Thread_Main_T)(void *);

class Thread {
public:
	Thread ( Thread_Main_T p_main ) : thread_main(p_main) {};
	~Thread () {};

	int run ( void *arg ) {
		return pthread_create(&pid,NULL,thread_main,arg);
	};

	void join () {
		pthread_join(pid,NULL);
	};
private:
	pthread_t pid;
	Thread_Main_T thread_main;
};

// ***********************************************************************
// Semaphore
// ***********************************************************************

class Semaphore {
public:
	Semaphore ( int p_initial, int p_max ) {
		sem_init(&sema,0,p_initial);
	};
	~Semaphore () {
		sem_destroy(&sema);
	};

	void P () { sem_wait(&sema); };
	void V () { sem_post(&sema); };

	int get_value () {
		int v;
		sem_getvalue(&sema,&v);
		return v;
	};
private:
	sem_t sema;
};

// ***********************************************************************
// SharedMemory
// ***********************************************************************

#include <stdio.h>

class SharedMemory {
public:
	SharedMemory ( const std::string &p_filename, int p_size )
	: filename(p_filename), len(p_size) {
		fd = shm_open(filename.c_str(),O_RDWR|O_CREAT,0x1ff);
		if (fd == -1) perror("shm_open");
		std::cout << "fd = " << fd << endl;
		ftruncate(fd,len);
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);	
	};

	SharedMemory ( const std::string &p_filename )
	: filename(p_filename) {
		struct stat buf;
		stat(filename.c_str(),&buf);
		len = buf.st_size;
		fd = shm_open(filename.c_str(),O_RDWR,0);
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
	};

	~SharedMemory () {
		munmap(addr,len);
		shm_unlink(filename.c_str());
	};

	char *base () const {
		return reinterpret_cast<char *>(addr);
	};

	int size () const {
		return len;
	};
private:
	std::string filename;
	int fd;
	void *addr;
	int len;
};


// ***********************************************************************
// MappedFile
// ***********************************************************************

class MappedFile {
public:
	MappedFile ( std::string filename ) {
		struct stat buf;
		stat(filename.c_str(),&buf);
		len = buf.st_size;
		fd = open(filename.c_str(),O_RDWR);
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
		
	};
	~MappedFile () {
		munmap(addr,len);
	};

	char *base () const {
		return reinterpret_cast<char *>(addr);
	};

	int size () const {
		return len;
	};
private:
	int fd;
	void *addr;
	int len;
};


